@@ -7,17 +7,18 @@
 // notice may not be copied, modified, or distributed except
 // according to those terms.
 
-use crate::parse::{Instruction, Spirv};
-use crate::structs;
-use crate::{spirv_search, TypesMeta};
+use crate::TypesMeta;
+use crate::{structs, RegisteredType};
 use proc_macro2::{Span, TokenStream};
-use spirv_headers::Decoration;
+use std::borrow::Cow;
+use std::collections::HashMap;
 use std::mem;
 use syn::Ident;
+use vulkano::spirv::{Decoration, Instruction, Spirv};
 
 /// Returns true if the document has specialization constants.
-pub fn has_specialization_constants(doc: &Spirv) -> bool {
-    for instruction in doc.instructions.iter() {
+pub fn has_specialization_constants(spirv: &Spirv) -> bool {
+    for instruction in spirv.iter_global() {
         match instruction {
             &Instruction::SpecConstantTrue { .. } => return true,
             &Instruction::SpecConstantFalse { .. } => return true,
@@ -32,11 +33,18 @@ pub fn has_specialization_constants(doc: &Spirv) -> bool {
 
 /// Writes the `SpecializationConstants` struct that contains the specialization constants and
 /// implements the `Default` and the `vulkano::pipeline::shader::SpecializationConstants` traits.
-pub(super) fn write_specialization_constants(doc: &Spirv, types_meta: &TypesMeta) -> TokenStream {
+pub(super) fn write_specialization_constants<'a>(
+    shader: &'a str,
+    spirv: &Spirv,
+    types_meta: &TypesMeta,
+    shared_constants: bool,
+    types_registry: &'a mut HashMap<String, RegisteredType>,
+) -> TokenStream {
     struct SpecConst {
         name: String,
         constant_id: u32,
         rust_ty: TokenStream,
+        rust_signature: Cow<'static, str>,
         rust_size: usize,
         rust_alignment: u32,
         default_value: TokenStream,
@@ -44,8 +52,8 @@ pub(super) fn write_specialization_constants(doc: &Spirv, types_meta: &TypesMeta
 
     let mut spec_consts = Vec::new();
 
-    for instruction in doc.instructions.iter() {
-        let (type_id, result_id, default_value) = match instruction {
+    for instruction in spirv.iter_global() {
+        let (result_type_id, result_id, default_value) = match instruction {
             &Instruction::SpecConstantTrue {
                 result_type_id,
                 result_id,
@@ -59,45 +67,71 @@ pub(super) fn write_specialization_constants(doc: &Spirv, types_meta: &TypesMeta
             &Instruction::SpecConstant {
                 result_type_id,
                 result_id,
-                ref data,
+                ref value,
             } => {
                 let def_val = quote! {
-                    unsafe {{ ::std::mem::transmute([ #( #data ),* ]) }}
+                    unsafe {{ ::std::mem::transmute([ #( #value ),* ]) }}
                 };
                 (result_type_id, result_id, def_val)
             }
             &Instruction::SpecConstantComposite {
                 result_type_id,
                 result_id,
-                ref data,
+                ref constituents,
             } => {
+                let constituents = constituents.iter().map(|&id| u32::from(id));
                 let def_val = quote! {
-                    unsafe {{ ::std::mem::transmute([ #( #data ),* ]) }}
+                    unsafe {{ ::std::mem::transmute([ #( #constituents ),* ]) }}
                 };
                 (result_type_id, result_id, def_val)
             }
             _ => continue,
         };
 
-        let (rust_ty, rust_size, rust_alignment) =
-            spec_const_type_from_id(doc, type_id, types_meta);
+        // Translate bool to u32
+        let (rust_ty, rust_signature, rust_size, rust_alignment) =
+            match spirv.id(result_type_id).instruction() {
+                Instruction::TypeBool { .. } => (
+                    quote! {u32},
+                    Cow::from("u32"),
+                    Some(mem::size_of::<u32>()),
+                    mem::align_of::<u32>(),
+                ),
+                _ => structs::type_from_id(shader, spirv, result_type_id, types_meta),
+            };
         let rust_size = rust_size.expect("Found runtime-sized specialization constant");
 
-        let constant_id = doc.get_decoration_params(result_id, Decoration::SpecId);
+        let id_info = spirv.id(result_id);
+
+        let constant_id = id_info
+            .iter_decoration()
+            .find_map(|instruction| match instruction {
+                Instruction::Decorate {
+                    decoration:
+                        Decoration::SpecId {
+                            specialization_constant_id,
+                        },
+                    ..
+                } => Some(*specialization_constant_id),
+                _ => None,
+            });
 
         if let Some(constant_id) = constant_id {
-            let constant_id = constant_id[0];
-
-            let mut name = spirv_search::name_from_id(doc, result_id);
-
-            if name == "__unnamed".to_owned() {
-                name = String::from(format!("constant_{}", constant_id));
-            }
+            let name = match id_info
+                .iter_name()
+                .find_map(|instruction| match instruction {
+                    Instruction::Name { name, .. } => Some(name.as_str()),
+                    _ => None,
+                }) {
+                Some(name) => name.to_owned(),
+                None => format!("constant_{}", constant_id),
+            };
 
             spec_consts.push(SpecConst {
                 name,
                 constant_id,
                 rust_ty,
+                rust_signature,
                 rust_size,
                 rust_alignment: rust_alignment as u32,
                 default_value,
@@ -105,6 +139,38 @@ pub(super) fn write_specialization_constants(doc: &Spirv, types_meta: &TypesMeta
         }
     }
 
+    let struct_name = Ident::new(
+        &format!(
+            "{}SpecializationConstants",
+            if shared_constants { "" } else { shader }
+        ),
+        Span::call_site(),
+    );
+
+    // For multi-constants mode registration mechanism skipped
+    if shared_constants {
+        let target_type = RegisteredType {
+            shader: shader.to_string(),
+            signature: spec_consts
+                .iter()
+                .map(|member| (member.name.to_string(), member.rust_signature.clone()))
+                .collect(),
+        };
+
+        let name = struct_name.to_string();
+
+        // Checking with Registry if this struct already registered by another shader, and if their
+        // signatures match.
+        if let Some(registered) = types_registry.get(name.as_str()) {
+            registered.assert_signatures(name.as_str(), &target_type);
+
+            // If the struct already registered and matches this one, skip duplicate.
+            return quote! {};
+        }
+
+        debug_assert!(types_registry.insert(name, target_type).is_none());
+    }
+
     let map_entries = {
         let mut map_entries = Vec::new();
         let mut curr_offset = 0;
@@ -143,19 +209,19 @@ pub(super) fn write_specialization_constants(doc: &Spirv, types_meta: &TypesMeta
         #[derive(Debug, Copy, Clone)]
         #[allow(non_snake_case)]
         #[repr(C)]
-        pub struct SpecializationConstants {
+        pub struct #struct_name {
             #( #struct_members ),*
         }
 
-        impl Default for SpecializationConstants {
-            fn default() -> SpecializationConstants {
-                SpecializationConstants {
+        impl Default for #struct_name {
+            fn default() -> #struct_name {
+                #struct_name {
                     #( #struct_member_defaults ),*
                 }
             }
         }
 
-        unsafe impl SpecConstsTrait for SpecializationConstants {
+        unsafe impl SpecConstsTrait for #struct_name {
             fn descriptors() -> &'static [SpecializationMapEntry] {
                 static DESCRIPTORS: [SpecializationMapEntry; #num_map_entries] = [
                     #( #map_entries ),*
@@ -165,25 +231,3 @@ pub(super) fn write_specialization_constants(doc: &Spirv, types_meta: &TypesMeta
         }
     }
 }
-
-// Wrapper around `type_from_id` that also handles booleans.
-fn spec_const_type_from_id(
-    doc: &Spirv,
-    searched: u32,
-    types_meta: &TypesMeta,
-) -> (TokenStream, Option<usize>, usize) {
-    for instruction in doc.instructions.iter() {
-        match instruction {
-            &Instruction::TypeBool { result_id } if result_id == searched => {
-                return (
-                    quote! {u32},
-                    Some(mem::size_of::<u32>()),
-                    mem::align_of::<u32>(),
-                );
-            }
-            _ => (),
-        }
-    }
-
-    structs::type_from_id(doc, searched, types_meta)
-}
